template = """
    <!doctype html>
    <html class="no-js" lang="">
        <head>
            <meta charset="utf-8">
            <meta http-equiv="x-ua-compatible" content="ie=edge">
            <title>{{ service_name | capitalize }}</title>
            <meta name="description" content="Landing">
            <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
            <link href="https://fonts.googleapis.com/css?family=Open+Sans:300,400,600,700" rel="stylesheet">
        </head>
        <style>
            html {
                font-family: sans-serif;
                -ms-text-size-adjust: 100%;
                -webkit-text-size-adjust: 100%;
                font-size: 10px;
                -webkit-tap-highlight-color: rgba(0, 0, 0, 0);
            }
            body {
                margin: 0;
                font-family: "Open Sans", sans-serif;
                font-size: 14px;
                line-height: 1.42857143;
                color: #333333;
                background-color: #fff;
            }
            a {
                background-color: transparent;
                color: #337ab7;
                text-decoration: none;
            }

            h1 {
                font-weight: 500;
                line-height: 1.1;
                color: inherit;
            }
            pre {
                background-color: #E8E8E8;
                border-style: solid;
                border-width: 1px;
            }
            .float-left {
                float: left;
            }
            .float-right {
                float: right;
            }
            .section {
                margin: 25px;
            }
            .section-health {
                padding-top: 25px;
                display: grid;
                grid-template-columns: 50% 50%;
            }
            .section-health > pre {
                height: 500px;
                overflow: scroll;
            }
            .text-center {
                text-align: center;
            }
            .no-margin {
                margin: 0px;
            }
            #env-button {
                margin: 5px 10px;
            }
        </style>
        <body>
            <h1 class="text-center">{{ service_name | capitalize }} - {{ version }}</h1>
            {%- if description -%}
            <h3 class="text-center">{{ description }}</h2>
            {%- endif -%}
            <div class="section section-health">
                <h2 class="no-margin"><a href="api/health">Health</a></h2>
                <div>
                    <h2 class="no-margin float-left"><a href="api/config">Config</a></h2>
                </div>
                <pre id="health">{{ health }}</pre>
                <pre id="config"><a id="env-button" class="float-right" href="data:text/plain;charset=utf-8,
                    {%- for item in env | sort -%}
                        {{ item | urlencode }}%0A
                    {%- endfor %}" download="{{ service_name }}_env"><button>Download Env</button></a>{{ config }}</pre>
            </div>
            {%- if homepage -%}
                <div class="section">
                    <h2><a href={{ homepage }}>Home Page</a></h2>
                </div>
            {%- endif -%}
            {%- if links -%}
            <div class="section">
                <h2>Links</h2>
                {%- for link in links | sort -%}
                <ul>
                    <li>
                        <a href="{{ links[link] }}">{{ link | replace("_", " ") | capitalize }}</a>
                    </li>
                </ul>
                {%- endfor -%}
            </div>
            {%- endif -%}
        </body>
    </html>
"""



"""
Path building.

"""
from microcosm.api import defaults


@defaults(
    prefix="/api",
)
class RoutePathBuilder:

    def __init__(self, graph):
        self.prefix = graph.config.build_route_path.prefix

    def __call__(self, path, prefix=None):
        """
        Generate a path for a route.

        :param path: the route path; must not be None
        :param prefix: overrides the default prefix if not None

        """
        return "/" + "/".join(filter(None, self.iter_path_parts(path, prefix)))

    def iter_path_parts(self, path, prefix):
        # use the provided prefix is not None
        if prefix is not None:
            yield prefix.strip("/")
        else:
            yield self.prefix.strip("/")

        # use the path
        yield path.strip("/")



from typing import Optional, Sequence

from marshmallow.fields import Field

from microcosm_flask.fields import EnumField
from microcosm_flask.swagger.parameters.base import ParameterBuilder


def is_int(value):
    try:
        int(value)
    except Exception:
        return False
    else:
        return True


class EnumParameterBuilder(ParameterBuilder):
    """
    Build an enum parameter.

    """
    def supports_field(self, field: Field) -> bool:
        return bool(getattr(field, "enum", None))

    def parse_format(self, field: Field) -> Optional[str]:
        if isinstance(field, EnumField):
            return "enum"
        return None

    def parse_type(self, field: Field) -> str:
        enum_values = self.parse_enum_values(field)

        if all((isinstance(enum_value, str) for enum_value in enum_values)):
            return "string"
        elif all((is_int(enum_value) for enum_value in enum_values)):
            return "integer"
        else:
            raise Exception(f"Cannot infer enum type for field: {field.name}")

    def parse_enum_values(self, field: Field) -> Sequence:
        enum = getattr(field, "enum", None)
        return [
            choice.value if field.by_value else choice.name  # type: ignore
            for choice in enum
        ]



from collections import namedtuple
from logging import Logger
from typing import Optional

from marshmallow import fields
from marshmallow.fields import Field
from microcosm_logging.decorators import logger

from microcosm_flask.fields import LanguageField, URIField
from microcosm_flask.swagger.parameters.base import ParameterBuilder


FieldInfo = namedtuple("FieldInfo", ["type", "format"])


FIELD_MAPPINGS = {
    LanguageField: FieldInfo("string", "language"),
    URIField: FieldInfo("string", "uri"),
    fields.Boolean: FieldInfo("boolean", None),
    fields.Date: FieldInfo("string", "date"),
    fields.DateTime: FieldInfo("string", "date-time"),
    fields.Decimal: FieldInfo("number", None),
    fields.Dict: FieldInfo("object", None),
    fields.Email: FieldInfo("string", "email"),
    fields.Field: FieldInfo("object", None),
    fields.Float: FieldInfo("number", "float"),
    fields.Integer: FieldInfo("integer", "int32"),
    fields.Method: FieldInfo("object", None),
    fields.Number: FieldInfo("number", None),
    fields.Raw: FieldInfo("object", None),
    fields.String: FieldInfo("string", None),
    fields.Time: FieldInfo("string", None),
    fields.URL: FieldInfo("string", "url"),
    fields.UUID: FieldInfo("string", "uuid"),
}


@logger
class DefaultParameterBuilder(ParameterBuilder):
    """
    Builds parameters using default field mappings.

    """
    logger: Logger

    def supports_field(self, field: Field) -> bool:
        try:
            FIELD_MAPPINGS[type(field)]
        except KeyError:
            self.logger.exception(
                f"No mapped swagger type for marshmallow field: {field}",
            )
            raise
        else:
            return True

    def parse_format(self, field: Field) -> Optional[str]:
        return FIELD_MAPPINGS[type(field)].format

    def parse_type(self, field: Field) -> Optional[str]:
        return FIELD_MAPPINGS[type(field)].type



from marshmallow.fields import Field, Nested

from microcosm_flask.naming import name_for
from microcosm_flask.swagger.naming import type_name
from microcosm_flask.swagger.parameters.base import ParameterBuilder


class NestedParameterBuilder(ParameterBuilder):
    """
    Builder parameters for nested fields.

    """
    def supports_field(self, field: Field) -> bool:
        return isinstance(field, Nested)

    def parse_ref(self, field: Field) -> str:
        """
        Parse the reference type for nested fields, if any.

        """
        ref_name = type_name(name_for(field.schema))  # type:ignore
        return f"#/definitions/{ref_name}"

    def parse_type(self, field: Field) -> None:
        return None



from flask import request
from jaeger_client.constants import TRACE_ID_HEADER
from microcosm.api import defaults
from microcosm.tracing import SPAN_NAME


X_REQUEST = "X-Request"
HEADER_PREFIXES = [X_REQUEST, TRACE_ID_HEADER]


def context_wrapper(include_header_prefixes):
    def retrieve_context():
        context = {
            header: value
            for header, value in request.headers.items()
            if any([
                header.lower().startswith(prefix.lower())
                for prefix in include_header_prefixes
            ])
        }
        span_name = f"{request.method} {request.path}"
        context[SPAN_NAME] = span_name
        return context
    return retrieve_context


@defaults(
    include_header_prefixes=HEADER_PREFIXES,
)
def configure_request_context(graph):
    """
    Configure the flask context function which controls what data you want to associate
    with your flask request context, e.g. headers, request body/response.

    Usage:
        graph.request_context()

    """
    include_header_prefixes = graph.config.request_context.include_header_prefixes
    return context_wrapper(include_header_prefixes)



from abc import ABC
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

from marshmallow import Schema
from marshmallow.fields import Field


ENTRY_POINT = "microcosm_flask.swagger.parameters"


class ParameterBuilder(ABC):
    """
    Plugin-aware swagger parameter builder.

    Discovers builder subclasses via the `microcosm_flask.swagger.parameters` entry point
    and delegates to the first compatible implementation.

    """
    def __init__(self, build_parameter: Callable[[Schema], Mapping[str, Any]]):
        self.build_parameter = build_parameter
        self.parsers = {
            "default": self.parse_default,
            "description": self.parse_description,
            "enum": self.parse_enum_values,
            "format": self.parse_format,
            "$ref": self.parse_ref,
            "type": self.parse_type,
            "items": self.parse_items,
            "x-nullable": self.parse_nullable,
        }

    def build(self, field: Field) -> Mapping[str, Any]:
        """
        Build a parameter.

        """
        return dict(self.iter_parsed_values(field))

    def supports_field(self, field: Field) -> bool:
        """
        Does this builder support this kind of field?

        """
        return False

    def iter_parsed_values(self, field: Field) -> Iterable[Tuple[str, Any]]:
        """
        Walk the dictionary of parsers and emit all non-null values.

        """
        for key, func in self.parsers.items():
            value = func(field)
            if not value:
                continue
            yield key, value

    def parse_default(self, field: Field) -> Any:
        """
        Parse the default value for the field, if any.

        """
        return field.default

    def parse_nullable(self, field: Field) -> Optional[bool]:
        if field.allow_none:
            return True
        return None

    def parse_description(self, field: Field) -> Optional[str]:
        """
        Parse the description for the field, if any.

        """
        return field.metadata.get("description")

    def parse_enum_values(self, field: Field) -> Optional[Sequence]:
        """
        Parse enumerated value for enum fields, if any.

        """
        return None

    def parse_format(self, field: Field) -> Optional[str]:
        """
        Parse the format for the field, if any.

        """
        return None

    def parse_items(self, field: Field) -> Optional[Mapping[str, Any]]:
        """
        Parse the child item type for list fields, if any.

        """
        return None

    def parse_ref(self, field: Field) -> Optional[str]:
        """
        Parse the reference type for nested fields, if any.

        """
        return None

    def parse_type(self, field: Field) -> Optional[str]:
        """
        Parse the type for the field, if any.

        """
        return None



from typing import Any, Mapping

from marshmallow.fields import Field, List

from microcosm_flask.swagger.parameters.base import ParameterBuilder


class ListParameterBuilder(ParameterBuilder):
    """
    Builder parameters for nested fields.

    """
    def supports_field(self, field: Field) -> bool:
        return isinstance(field, List)

    def parse_items(self, field: Field) -> Mapping[str, Any]:
        """
        Parse the child item type for list fields, if any.

        """
        return self.build_parameter(field.inner)  # type: ignore

    def parse_type(self, field: Field) -> str:
        return "array"



"""
An RFC-1766 and xsd:language compatible field.

"""
from re import match

from marshmallow.fields import String


# taken from: http://books.xmlschemata.org/relaxng/ch19-77191.html
LANGUAGE_PATTERN = r"^([a-zA-Z]{2}|[iI]-[a-zA-Z]+|[xX]-[a-zA-Z]{1,8})(-[a-zA-Z]{1,8})*$"


class LanguageField(String):
    default_error_messages = dict(
        invalid_language="Not an RFC-1766 language",
    )

    def _validated(self, value):
        if not match(LANGUAGE_PATTERN, value):
            self.fail("invalid_language")
        return value

    def _serialize(self, value, attr, obj, **kwargs):
        validated = str(self._validated(value)) if value is not None else None
        return super(LanguageField, self)._serialize(validated, attr, obj)

    def _deserialize(self, value, attr, data, **kwargs):
        return self._validated(value)



from microcosm_flask.conventions.crud_adapter import CRUDStoreAdapter


class EncryptableCRUDStoreAdapter(CRUDStoreAdapter):
    """
    Adapt the CRUD conventions callbacks to the `EncryptableStore` interface.

    """

    def update_and_reencrypt(self, **kwargs):
        """
        Support re-encryption by enforcing that every update triggers a
        new encryption call, even if the the original call does not update
        the encrypted field.

        """
        encrypted_field_name = self.store.model_class.__plaintext__

        id_ = kwargs[self.identifier_key]
        current_model = self.store.retrieve(id_)
        current_value = current_model.plaintext

        null_update = (
            # Check if the update is for the encrypted field, and if it's explicitly set to null
            encrypted_field_name in kwargs
            and kwargs.get(encrypted_field_name) is None
        )
        new_value = kwargs.pop(self.store.model_class.__plaintext__, None)
        use_new_value = new_value is not None or null_update

        updated_value = new_value if use_new_value else current_value

        model_kwargs = {
            self.identifier_key: id_,
            encrypted_field_name: updated_value,
            **kwargs,
        }

        return super().update(
            **model_kwargs,
        )



"""
Saved search convention.

"""
from functools import wraps

from inflection import pluralize

from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import dump_response_data, load_request_data, merge_data
from microcosm_flask.conventions.registry import request, response
from microcosm_flask.operations import Operation
from microcosm_flask.paging import OffsetLimitPage, identity


class SavedSearchConvention(Convention):

    @property
    def page_cls(self):
        return OffsetLimitPage

    def configure_savedsearch(self, ns, definition):
        """
        Register a saved search endpoint.

        The definition's func should be a search function, which must:
        - accept kwargs for the request data
        - return a tuple of (items, count) where count is the total number of items
          available (in the case of pagination)

        The definition's request_schema will be used to process query string arguments.

        :param ns: the namespace
        :param definition: the endpoint definition

        """
        paginated_list_schema = self.page_cls.make_paginated_list_schema_class(
            ns,
            definition.response_schema,
        )()

        @self.add_route(ns.collection_path, Operation.SavedSearch, ns)
        @request(definition.request_schema)
        @response(paginated_list_schema)
        @wraps(definition.func)
        def saved_search(**path_data):
            request_data = load_request_data(definition.request_schema)
            page = self.page_cls.from_dict(request_data)
            request_data.update(page.to_dict(func=identity))
            result = definition.func(**merge_data(path_data, request_data))
            response_data, headers = page.to_paginated_list(result, ns, Operation.SavedSearch)
            definition.header_func(headers, response_data)
            response_format = self.negotiate_response_content(definition.response_formats)
            return dump_response_data(
                paginated_list_schema,
                response_data,
                headers=headers,
                response_format=response_format,
            )

        saved_search.__doc__ = "Persist and return the search results of {}".format(pluralize(ns.subject_name))


def configure_saved_search(graph, ns, mappings):
    convention = SavedSearchConvention(graph)
    convention.configure(ns, mappings)
